/*
 * Copyright (c) 2013, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== empty.c ========
 */
/* XDCtools Header files */
#include <xdc/std.h>
#include <xdc/cfg/global.h>
#include <xdc/runtime/System.h>

/* BIOS Header files */
#include <ti/sysbios/BIOS.h>

/* TI-RTOS Header files */
#include <ti/drivers/GPIO.h>
#include <ti/drivers/SPI.h>

/* Example/Board Header files */
#include "Board.h"

#include "AT45DB.h"
#include <string.h>

#define SPI_MSG_LENGTH 1

UChar textarray[AT45DB161_PAGE_SIZE] = \
"1234567890123456789012345678901234567890123456789012345678901234567890\n"
"         1         2         3         4         5         6         7\n"
"This is some text to be inserted into the AT45DB device.              \n"
"What is says does not really matter so I'm going to list the          \n"
"undisputed chess world champions: Steinitz, Lasker, Capablanca,       \n"
"Alekhine, Euwe, Botvinnik, Smyslov, Tal, Petrosian, Spassky, Fischer, \n"
"Karpov, Kasparov....Kramnik, Anand, Carlsen.                          \n";

UChar readarray[AT45DB161_PAGE_SIZE];

/*
 *  ======== writeBufferTest ========
 *  Testing out the erasing of a page.
 *  1. write data1 to the buffer
 *  2. write data2 to the buffer
 *  3. read back the buffer and validate
 */
void writeBufferTest(AT45DB_Handle at45dbHandle, uint32_t page,
                     unsigned char *data1, uint32_t data_size1,
                     unsigned char *data2, uint32_t data_size2)
{
    AT45DB_Transaction at45dbTransaction;
    int i, j;

    /* Write data1 */
    at45dbTransaction.data      = data1;
    at45dbTransaction.data_size = data_size1;
    at45dbTransaction.byte      = 0;

    if (!AT45DB_writeBuffer(at45dbHandle, &at45dbTransaction)) {
       /* Print contents of slave receive buffer */
      System_abort("Unsuccessful SPI transfer");
    }

    /* Write data2 */
    at45dbTransaction.data      = data2;
    at45dbTransaction.data_size = data_size2;
    at45dbTransaction.byte      = data_size1;

    if (!AT45DB_writeBuffer(at45dbHandle, &at45dbTransaction)) {
       /* Print contents of slave receive buffer */
      System_abort("Unsuccessful SPI transfer");
    }

    /* Read and validate */
    at45dbTransaction.data      = readarray;
    at45dbTransaction.data_size = data_size1 + data_size2;
    at45dbTransaction.byte      = 0;
    if (AT45DB_readBuffer(at45dbHandle, &at45dbTransaction)) {
        for (i = 0; i < data_size1; i++) {
            if (readarray[i] != data1[i]) {
                System_printf("validation failed written 0x%x, read 0x%x",
                              data1[i], readarray[i]);
            	System_abort("");
            }
        }
        for (j = 0; i < data_size2; j++, i++) {
            if (readarray[i] != data2[j]) {
                System_printf("validation failed written 0x%x, read 0x%x",
                              data2[j], readarray[i]);
            	System_abort("");
            }
        }
    }
    else {
       /* Print contents of slave receive buffer */
      System_abort("Unsuccessful SPI transfer");
    }
}

/*
 *  ======== writeTest ========
 *  Testing out the erasing of a page.
 *  1. write into the page
 *  2. read the page and validate that it was what was written
 */
void writeTest(AT45DB_Handle at45dbHandle,
               uint32_t page, uint32_t byte,
               unsigned char *data, uint32_t data_size)
{
    AT45DB_Transaction at45dbTransaction;
    int i;

    at45dbTransaction.data      = data;
    at45dbTransaction.data_size = data_size;
    at45dbTransaction.byte      = byte;

    /* Write data */
    if (!AT45DB_write(at45dbHandle, &at45dbTransaction, page)) {
       /* Print contents of slave receive buffer */
      System_abort("Unsuccessful SPI transfer");
    }

    /* Read and validate */
    at45dbTransaction.data = readarray;
    at45dbTransaction.data_size = data_size;
    at45dbTransaction.byte      = byte;
    if (AT45DB_read(at45dbHandle, &at45dbTransaction, page)) {
        for (i = 0; i < data_size; i++) {
            if (readarray[i] != data[i]) {
                System_printf("validation failed written 0x%x, read 0x%x",
                              data[i], readarray[i]);
            	System_abort("");
            }
        }
    }
    else {
        System_abort("Unsuccessful SPI transfer");
    }
}

/*
 *  ======== writeBufferToPageTest ========
 *  Testing out the erasing of a page.
 *  1. write two separate known strings to the buffer
 *  2. push the buffer to the page
 *  3. read back the buffer and validate
 */
void writeBufferToPageTest(AT45DB_Handle at45dbHandle, uint32_t page)
{
    AT45DB_Transaction at45dbTransaction;
    int i;

    /* Write to the buffer */
    writeBufferTest(at45dbHandle, page, &(textarray[70]), 10, &(textarray[80]), 16);

    /* Push to the page */
    AT45DB_writeBufferToPage(at45dbHandle, page);

    /* Read and validate */
    at45dbTransaction.data      = readarray;
    at45dbTransaction.data_size = 10 + 16;
    at45dbTransaction.byte      = 0;
    if (AT45DB_read(at45dbHandle, &at45dbTransaction, page)) {
        for (i = 0; i < 10 + 16; i++) {
            if (readarray[i] != textarray[70 + i]) {
                System_printf("failed to erase to 0xff (%d)", readarray[i]);
                System_abort("");
            }
        }
    }
    else {
        System_abort("Unsuccessful SPI transfer");
    }
}

/*
 *  ======== eraseTest ========
 *  Testing out the erasing of a page.
 *  1. write a known string into the page (via writeTest)
 *  2. erase the page
 *  3. read the page and confirm that it is erased
 */
void eraseTest(AT45DB_Handle at45dbHandle, uint32_t page)
{
    AT45DB_Transaction at45dbTransaction;
    int i;

    /* Write and validate known string */
    writeTest(at45dbHandle, page, 0, textarray, sizeof(readarray));

    /* Erase */
    AT45DB_erasePage(at45dbHandle, page);

    /* Read and validate */
    at45dbTransaction.data      = readarray;
    at45dbTransaction.data_size = AT45DB161_PAGE_SIZE;
    at45dbTransaction.byte      = 0;
    if (AT45DB_read(at45dbHandle, &at45dbTransaction, page)) {
        for (i = 0; i < AT45DB161_PAGE_SIZE; i++) {
            if (readarray[i] != 0xff) {
                System_printf("failed to erase to 0xff (%d)", readarray[i]);
                System_abort("");
            }
        }
    }
    else {
        System_abort("Unsuccessful SPI transfer");
    }
}

/*
 *  ======== myTaskFxn ========
 *  Testing function
 */
Void myTaskFxn (UArg arg0, UArg arg1)
{
    SPI_Params spiParams;
    SPI_Handle spiHandle;
    AT45DB_Params at45dbParams;
    AT45DB_Object obj;
    AT45DB_Handle at45dbHandle;
    volatile unsigned char ready;

    /* Open the SPI port */
    SPI_Params_init(&spiParams);    
    spiHandle = SPI_open(Board_SPI0, &spiParams);
    
    /* Construct the AT45DB handle */
    AT45DB_Params_init(&at45dbParams);
    at45dbHandle = AT45DB_construct(&obj, spiHandle, Board_CS, &at45dbParams);
    if (at45dbHandle == NULL) {
        System_abort("AT45DB_construct failed");
    }

    writeTest(at45dbHandle, 0, 0, textarray, sizeof(readarray));
    writeTest(at45dbHandle, 1, 0, textarray, sizeof(readarray));
    writeTest(at45dbHandle, 2, 400, textarray, 100);
    writeTest(at45dbHandle, 6, 70, textarray, 75);

    writeBufferTest(at45dbHandle, 1, &(textarray[0]), 12, &(textarray[12]), 16);
    
    eraseTest(at45dbHandle, 0);
    eraseTest(at45dbHandle, 1);
    eraseTest(at45dbHandle, 2);
    
    writeBufferToPageTest(at45dbHandle, 3);
    writeBufferToPageTest(at45dbHandle, 4);

    System_printf("All tests passed! Calling BIOS_exit now. Bye...");
    BIOS_exit(0);
}

/*
 *  ======== main ========
 */
Int main(Void)
{
    /* Call board init functions. */
    Board_initGeneral();
    Board_initGPIO();
    Board_initSPI();

    /* Turn on user LED */
    GPIO_write(Board_LED0, Board_LED_OFF);

    System_printf("Starting the example\nSystem provider is set to SysMin. "
                  "Halt the target and use ROV to view output.\n");
    /* SysMin will only print to the console when you call flush or exit */
    System_flush();

    /* Start BIOS */
    BIOS_start();

    return (0);
}
